/* 
 * This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Author: Sascha Dietrich
 *
 * Created on 20. April 2012, 09:25
 */

#include <cstdlib>
#include <string>
#include <iostream>
#include <sstream>
#include <fstream>
#include <vector>
#include <set>
#include "read.hpp"
#include <boost/regex.hpp>
#include <boost/algorithm/string.hpp>
#include <map>
#include <boost/iostreams/filtering_streambuf.hpp>
#include <boost/iostreams/copy.hpp>
#include <boost/iostreams/filter/zlib.hpp>
#include <boost/multiprecision/cpp_int.hpp>

using namespace std;
namespace mp = boost::multiprecision; 

string VERSION = "1.01";
/*
 * 
 */

void setparams(map<string,string>* p_map, char** args, int argc) {
    string allcmds = "";
    int i;
    for (i = 1; i < argc; i++) {
        allcmds = allcmds + args[i] + " ";
    }
    
    string::iterator sit = allcmds.begin();
    int state = -1; //0 new param, 1 new comes value, -1 waiting for new param
    string param = "";
    string value = "";
    for (sit = allcmds.begin(); sit != allcmds.end(); sit++) {
        if (state == -1) {
            if (*sit == '-') {
                //new param
                state = 0;
                continue;
            }
        }
        
        if (state == 1) {
            if (*sit == ' ') {
                state = -1;
                (*p_map)[param] = value;
                param = "";
                value = "";
                continue;
            } else {
                value = value + *sit;
            }
        }
         
        if (state == 0) {
            //grabbing param
            if (*sit == ' ') {
                state = 1;
                continue;
            }
            else {
                param = param + *sit;
            }
        }

    }
    if (state == 1) { //exiting the loop when we should expect a parameter. That's a single param without value like -h
        (*p_map)[param] = value;
    }
}

vector<string> split(string input, char delimiter) {
    stringstream ss(input);
    string subString;
    vector<string> splitString;
    while(getline(ss, subString, delimiter)) {
        splitString.push_back(subString);
    }
    return splitString;
}

vector<string> splitCIGAR(string input) {
    vector<string> splitCigar;
    string curNum;
    
    for (string::iterator sit = input.begin(); sit < input.end(); ++sit){
        if (isalpha(*sit)) {
            splitCigar.push_back(curNum+=*sit);
            curNum = "";
        }
        else {
            curNum += *sit;
        }
    }
    return splitCigar;
}

int getTotalLenFromCIGAR(string input, bool noclip) {
    vector<string> splitCig = splitCIGAR(input);
    int matchlen = 0;
    for (vector<string>::iterator vit = splitCig.begin(); vit != splitCig.end(); vit++) {
        string curCig = *vit;
        stringstream curLenSS(curCig);
        size_t found;
        if (!noclip) {
            found = curCig.find("H");
            if (found != string::npos) {
                continue;
            }
            found = curCig.find("S");
            if (found != string::npos) {
                continue;
            }
        }
        int curLen;
        curLenSS >> curLen;
        matchlen = matchlen + curLen;
    }
    return matchlen;
}

int getMatchCountFromCIGAR(string input) {
    vector<string> splitCig = splitCIGAR(input);
    int matchlen = 0;
    for (vector<string>::iterator vit = splitCig.begin(); vit != splitCig.end(); vit++) {
        size_t found;
        string curCig = *vit;
        found = curCig.find("M");
        if (found != string::npos) {
            curCig.replace(curCig.find("M"), 1, "");
            stringstream curLenSS(curCig);
            int curLen;
            curLenSS >> curLen;
            matchlen = matchlen + curLen;
        }
        
        found = curCig.find("D");
        if (found != string::npos) {
            curCig.replace(curCig.find("D"), 1, "");
            stringstream curLenSS(curCig);
            int curLen;
            curLenSS >> curLen;
            matchlen = matchlen + curLen;
        }
        
        
    }
    return matchlen;
}

void printHelp() {
    cout  << endl<<  "SAMtoTDS version: " << VERSION << endl
            << "usage:" << endl <<endl
            << "samtotds -i <SAMFile> -o <Experiment Tag> -p <percent identity cutoff>" << endl
            << "e.g.: samtotds -i my.sam -o projectname -p 95 for 95% or better sequence identity." << endl
            << "-p: sets the percent identity limit. Calculate percent identity based on CIGAR strings and only map reads that lie above this threshold." << endl
            << "If you do not wish the filter by percent identity (thus trust the mapper completely) you can enter 0 here." << endl << endl
            << "Optional:" << endl
            << "-reads: put single read mappings in tds. Greatly increases file size but provides single read mapping information. Not used by TraV directly" << endl
            << "and not needed in normal usage." << endl
            << "-clip: use clipping information, don't treat them as mismatched positions." << endl << endl;
}


int main(int argc, char** argv) {
    map<string,string> params;
    setparams(&params, argv,argc);
    //map<string, map<string,Read> > contigReadMap;
    bool noclip = true;
    if (params.find("clip") != params.end()) noclip = false;
    
    map<string, int> repSizes;
    map<string,Read> readMap;
   
    if (params.find("h") != params.end()) {
        printHelp();
        exit(0);
    }
    
    if (params.find("i") == params.end())  {
        printHelp();
        exit(0);
    }

    if (params.find("o") == params.end()) {
        printHelp();
        exit(0);
    }
    if (params.find("p") == params.end()) {
        printHelp();
        exit(0);
    }
    
    
    int minpercIdent;
    if (params.find("p") != params.end()) {
        stringstream ss(params["p"]);
        ss >> minpercIdent;
    }
    string projName = params["o"];
    
    ifstream readFile;
    string l;
    readFile.open(params["i"].c_str(), ios::in | ios::binary);
    
    Read newRead;
    Read *p_newRead = &newRead; 
    //set<string> readList
    cout << "Processing read mapping from SAM" << endl;
    bool foundHeader = false;
    
    if (readFile.is_open()) {
        while(readFile.good()) {
            getline(readFile, l);
            
            string::iterator seqIt = l.begin();
            if (*seqIt == '@') {
                foundHeader = true;
                //vector<string> splitL = split(l, '\t');
                boost::regex getSeqName("@SQ\tSN:(.+)\tLN:(.+)");
                boost::smatch getSeqNameMat;
                if (boost::regex_match(l, getSeqNameMat, getSeqName, boost::match_extra)) {
                   int length;
                   stringstream lenss(getSeqNameMat[2]);
                   lenss >> length;
                   cout << getSeqNameMat[1] << " " << length << endl;
                   repSizes[getSeqNameMat[1]] = length;
                }
                continue;
            }
            
            if (foundHeader == false) {
                cout << "Your SAM file does not contain a header!" << endl;
                cout << "Unfortunately samtotds requires them." << endl << endl;
                cout << "You need to add a @SQ line for each replicon at the beginning in your sam file" << endl;
                cout << "e.g. if a sequence called \"plasmid\" was present and it is 75000 bases long:" << endl;
                cout << "@SQ    SN:plasmid      LN:75000" << endl;
                cout << "IMPORTANT: those are TAB characters, not spaces" << endl << endl;
                exit(1);
            }
            vector<string> splitL = split(l,'\t');
            if (splitL.size() < 10) {
                continue;
            }
            //readList.insert(splitL[0]);
            
            stringstream bitmaskss(splitL[1]);
            int bitmask;
            bitmaskss >> bitmask;
           
            
            new(p_newRead) Read();
            //readCount++;
            
            p_newRead->name = splitL[0];
            p_newRead->mapped = false;
            p_newRead->unique = true;
            p_newRead->contig = splitL[2];
            p_newRead->mapType = 0;
            
            if ((bitmask & 16) ==  16) {
                p_newRead->orient = 1;
            }
            else {
                p_newRead->orient = 0;
            }
            //cout << "readOrientVal:" << p_newRead->orient << endl;
            
            if ((bitmask & 4) == 4) {
                //output unmapped
                //unmapped++;
                p_newRead->mapType = 0;
                readMap[p_newRead->name] = *p_newRead;
                continue;
            }
            
            //multimap with bitflag
            if ((bitmask & 256) == 256) {
                p_newRead->mapped = false;
                p_newRead->mapType = 2;
                //readMap[p_newRead->name] = *p_newRead;
                //unmapped++;
                //continue;
            }
            
            stringstream startss(splitL[3]);
            startss >> p_newRead->start;

            
            stringstream cigarss(splitL[5]);
            string cigar;
            cigarss >> cigar;
           
            int matchlen = getMatchCountFromCIGAR(cigar);
            int len = getTotalLenFromCIGAR(cigar, noclip);
            
            //vector<string> cigarStrings = splitCIGAR(cigar);

            //substract NM count from matchlen to account mismatches
            if (l.find("NM:") != string::npos) {
                boost::regex getMismatches(".+NM:i:([\\-0-9]+).+");
                boost::smatch getMismatchesMat;
                if (boost::regex_match(l, getMismatchesMat, getMismatches, boost::match_extra)) {
                    int mmCount;
                    stringstream mmcss(getMismatchesMat[1]);
                    mmcss >> mmCount;
                    matchlen = matchlen - mmCount;
                }
            }
            
            float percMatch = (matchlen / (float) len) * 100;
            //cout << percMatch << endl;
            if (percMatch >= (float)minpercIdent ) {
                p_newRead->mapped = true;
                p_newRead->bitScore = percMatch;
                p_newRead->mapType = 1;
            }
            else {
                p_newRead->mapped = false;
                p_newRead->mapType = 0;
                if (readMap.find(p_newRead->name) == readMap.end()) {
                    readMap[p_newRead->name] = *p_newRead;
                }
                //unmapped++;
                continue;
            }
            
            p_newRead->stop = p_newRead->start + len - 1;
            //min length cutoff check
            
            //BWA multimap with XA: tag //special branch
            if (l.find("XA:") != string::npos) {
                boost::regex getSeqName(".+XA:Z:(.+);");
                boost::smatch getSeqNameMat;

                if (boost::regex_match(l, getSeqNameMat, getSeqName, boost::match_extra)) {
                    //cout << getSeqNameMat[1] << endl;
                    vector<string> splitXA = split(getSeqNameMat[1], ';');
                    for (vector<string>::iterator vit = splitXA.begin(); vit != splitXA.end(); vit++) {
                        vector<string> splitHit = split(*vit, ',');
                        int matchlen = getMatchCountFromCIGAR(splitHit[2]);
                        int len = getTotalLenFromCIGAR(splitHit[2], noclip);

                        int mmCount;
                        stringstream mmcss(splitHit[3]);
                        mmcss >> mmCount;
                        matchlen = matchlen - mmCount;
                        
                        float percMatch = (matchlen / (float) len) * 100;
                        if (p_newRead->bitScore == percMatch) {
                            p_newRead->unique = false;
                            p_newRead->mapType = 2;
                        }
                    }
                }
                
                readMap[p_newRead->name] = *p_newRead;
                //unmapped++;
                continue;
            }
            
            
            if (readMap.find(p_newRead->name) == readMap.end()) {
                //new read, just add
                readMap[p_newRead->name] = *p_newRead;
            }
            else {
                //read already in the readMap, let's check mapped and percIdentity
                if (readMap[p_newRead->name].mapType != 1) { //Hey, the one in Readmap isn't mapped yet
                    readMap[p_newRead->name] = *p_newRead;
                }
                else { //read in map is a mapped one, check bitscore if they're equal or one is better than the other
                    if (readMap[p_newRead->name].bitScore == p_newRead->bitScore) {
                        //equally good, make it multimapped
                        readMap[p_newRead->name].unique = false;
                        readMap[p_newRead->name].mapType = 2;
                    }
                    else if (readMap[p_newRead->name].bitScore < p_newRead->bitScore) {
                        //old one has worse percIdentity, replace
                        readMap[p_newRead->name] = *p_newRead;
                    }
                }
            }
        }
    }
    else {
        cout << "Could not open SAM file:" << endl;
        cout << params["i"].c_str() << endl;
        exit(1);
    }
    
    cout << "Read processing done" << endl;
    cout << "create covStrings" << endl;
    
    map<string,int*> covLeadMap;
    map<string,int*> covLagMap;
    //map<string, vector<int> > repLeadBMap;
    //map<string, vector<int> > repLagBMap;
   
    map<string, int*> mappedReads;

    //create covStrings here
    //0 init all covStrings for each contig
    
    for (map<string,int>::iterator crepIt = repSizes.begin(); crepIt != repSizes.end(); ++crepIt) {
        int curRepSize = repSizes[crepIt->first];
        //prep ints for mapped read counts per contig, used later
        int* p_mappedRead = (int*) malloc(sizeof(int));
        *p_mappedRead = 0;
        mappedReads[crepIt->first] = p_mappedRead;
        
        int* p_curCovLead = (int*)malloc((curRepSize+1)*sizeof(int));
        int* p_curCovLag = (int*)malloc((curRepSize+1)*sizeof(int));
        
        for (int i = 0; i <= curRepSize; i++) {
            p_curCovLead[i] = 0;
            p_curCovLag[i] = 0;
        }
        covLeadMap[crepIt->first] = p_curCovLead;
        covLagMap[crepIt->first] = p_curCovLag;
    }
    
    //now count up covs
    int totalMapped = 0;
    int multihits = 0;
    int unmapped = 0;
    int readCount = 0;
    for (map<string, Read>::iterator readIt = readMap.begin(); readIt != readMap.end(); ++readIt) {
        readCount++;
        if (readIt->second.mapType == 1) {
            totalMapped++;
            //cout << "readOrientVal:" << readIt->second.orient << endl;
            if (readIt->second.orient == 0) {
                
                for (int i = readIt->second.start; i <= readIt->second.stop; i++) {
                    if (i >=  repSizes[readIt->second.contig]) continue;
                    
                    covLeadMap[readIt->second.contig][i]++;
                }
            } else {
                for (int i = readIt->second.start; i <= readIt->second.stop; i++) {
                    if (i >=  repSizes[readIt->second.contig]) continue;
                    covLagMap[readIt->second.contig][i]++;
                }
            }
            (*mappedReads[readIt->second.contig])++;
        }
        if (readIt->second.mapType == 0) unmapped++;
        if (readIt->second.mapType == 2) multihits++; 
    }
    cout << "covstrings done" << endl;

    //calc totalBaseCov for experiment
    //mp::int1024_t
    mp::int512_t totalBaseCov = 0;
    for (map<string, int>::iterator crepIt = repSizes.begin(); crepIt != repSizes.end(); ++crepIt) {
        for (int i = 0; i <= repSizes[crepIt->first]; ++i) {
            totalBaseCov = totalBaseCov + covLeadMap[crepIt->first][i];
            totalBaseCov = totalBaseCov + covLagMap[crepIt->first][i];
        }
    }
    
    //Write TDS files
    for (map<string, int>::iterator crepIt = repSizes.begin(); crepIt != repSizes.end(); ++crepIt) {
        string tdsTmpFile = projName + "_" + crepIt->first + ".tds";
        //cout << ">" << tdsTmpFile.c_str() << "<" << endl;

        //tdsOut.open(tdsTmpFile.c_str());
        
        FILE *tdsOut;
        tdsOut = fopen(tdsTmpFile.c_str(), "w");
        if (tdsOut == NULL) {
            cout << "Failed to open file " << tdsTmpFile << endl;
            exit(1);
        }
        
        fprintf(tdsOut, "%d\n",*mappedReads[crepIt->first]);
        fprintf(tdsOut, "%d\n", (unmapped + multihits));
        fprintf(tdsOut, "%s\n", totalBaseCov.str().c_str());
        fprintf(tdsOut, "%s", "\\\\COV\n");
        for (int i = 0; i <= repSizes[crepIt->first]; ++i) {
            fprintf(tdsOut, "%d\t%d\n", covLeadMap[crepIt->first][i] ,(covLagMap[crepIt->first])[i]);
        }
        string orient;
        fprintf(tdsOut, "%s", "\\\\READS\n");
        //print out read coords for reads belonging to our contig
        if (params.find("reads") != params.end()) {
            for (map<string, Read>::iterator readIt = readMap.begin(); readIt != readMap.end(); ++readIt) {
                if (readIt->second.mapType == 1) {
                    if (strcmp(readIt->second.contig.c_str(), crepIt->first.c_str()) == 0) {
                        if (readIt->second.orient == 0) {
                            orient = "+";
                        } else {
                            orient = "-";
                        }
                        string name = readIt->second.name;

                        fprintf(tdsOut, "%s\t%d%s%d\n", name.c_str(), readIt->second.start, orient.c_str(), readIt->second.stop);
                    }
                }
            }
        }
        fflush(tdsOut);
        /*string gzipexec = "gzip " + tdsTmpFile + " -S .tds";
        system(gzipexec.c_str());*/
    }
    
    //multiout
    string multiFile = projName + "_multihits";
    ofstream multiOut(multiFile.c_str());
    
    string mappedFile = projName + "_mappedReads";
    ofstream mappedOut(mappedFile.c_str());
    
    string unmappedFile = projName + "_unmappedReads";
    ofstream unmappedOut(unmappedFile.c_str());
    
    for (map<string, Read>::iterator readIt = readMap.begin(); readIt != readMap.end(); ++readIt) {
        if (readIt->second.mapType == 0) {
            unmappedOut << readIt->second.name << endl;
        }
        
        if (readIt->second.mapType == 1) {
            mappedOut << readIt->second.name << endl;
        }
        
        if (readIt->second.mapType == 2) {
            multiOut << readIt->second.name << endl;
        }
    }   
    

    mappedOut.flush();
    mappedOut.close();
    unmappedOut.flush();
    unmappedOut.close();
    multiOut.flush();
    multiOut.close();
    
    cout << "TDS done" << endl << endl;
    cout << "Of " << readCount << " reads:" << endl;
    cout << "total mapped: " << totalMapped << endl;
    cout << "unmapped: " <<  unmapped << endl;
    cout << "multimap: " << multihits << endl; 

    readMap.clear();
    return 0;
}

